import datetime
import json
from typing import Any, Dict, List, Optional

from dbcat.catalog import Catalog, CatSchema, CatSource, CatTable
from dbcat.catalog.models import PiiTypes

from piicatcher.generators import column_generator


# Ref: https://stackoverflow.com/questions/24481852/serialising-an-enum-member-to-json
class PiiTypeEncoder(json.JSONEncoder):
    # pylint: disable=method-hidden
    def default(self, obj):
        if type(obj) == PiiTypes:
            return {"__enum__": str(obj)}
        return json.JSONEncoder.default(self, obj)


def output_dict(
    catalog: Catalog,
    source: CatSource,
    list_all: bool = False,
    last_run: datetime.datetime = None,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Dict[Any, Any]:
    current_schema: Optional[CatSchema] = None
    current_table: Optional[CatTable] = None

    source_dict = {"name": source.name, "schemata": []}
    schema_dict = {"name": "", "tables": []}

    table_dict = {"name": "", "columns": []}
    for schema, table, column in column_generator(
        catalog=catalog,
        source=source,
        last_run=last_run,
        exclude_schema_regex_str=exclude_schema_regex,
        include_schema_regex_str=include_schema_regex,
        exclude_table_regex_str=exclude_table_regex,
        include_table_regex_str=include_table_regex,
    ):
        if current_schema is None or schema != current_schema:
            if current_schema is not None:
                if len(table_dict["columns"]) > 0 or list_all:
                    schema_dict["tables"].append(table_dict)  # type: ignore
                if len(schema_dict["tables"]) > 0 or list_all:
                    source_dict["schemata"].append(schema_dict)
            current_schema = schema
            schema_dict = {"name": schema.name, "tables": []}
            current_table = None

        if current_table is None or table != current_table:
            if current_table is not None:
                if len(table_dict["columns"]) > 0 or list_all:
                    schema_dict["tables"].append(table_dict)  # type: ignore
            current_table = table
            table_dict = {"name": table.name, "columns": []}
        if column.pii_type is not None or list_all:
            table_dict["columns"].append(  # type: ignore
                {
                    "name": column.name,
                    "data_type": column.data_type,
                    "sort_order": column.sort_order,
                    "pii_type": column.pii_type,
                    "pii_plugin": column.pii_plugin,
                }
            )
    if len(table_dict["columns"]) > 0 or list_all:
        schema_dict["tables"].append(table_dict)  # type: ignore
    if len(schema_dict["tables"]) > 0 or list_all:
        source_dict["schemata"].append(schema_dict)

    return source_dict if len(source_dict["schemata"]) > 0 or list_all else {}


def output_tabular(
    catalog: Catalog,
    source: CatSource,
    list_all: bool = False,
    last_run: datetime.datetime = None,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> List[Any]:
    tabular = []

    for schema, table, column in column_generator(
        catalog=catalog,
        source=source,
        last_run=last_run,
        exclude_schema_regex_str=exclude_schema_regex,
        include_schema_regex_str=include_schema_regex,
        exclude_table_regex_str=exclude_table_regex,
        include_table_regex_str=include_table_regex,
    ):
        if list_all or column.pii_type is not None:
            tabular.append(
                [
                    schema.name,
                    table.name,
                    column.name,
                    str(column.pii_type),
                    column.pii_plugin,
                ]
            )

    return tabular



import json
from pathlib import Path
from typing import List, Optional

import typer
from tabulate import tabulate

from piicatcher.api import (
    OutputFormat,
    ScanTypeEnum,
    scan_athena,
    scan_mysql,
    scan_postgresql,
    scan_redshift,
    scan_snowflake,
    scan_sqlite,
)
from piicatcher.app_state import app_state
from piicatcher.generators import NoMatchesError
from piicatcher.output import PiiTypeEncoder

app = typer.Typer()


schema_help_text = """
Scan only schemas matching schema; When this option is not specified, all
non-system schemas in the target database will be dumped. Multiple schemas can
be selected by writing multiple --include switches. Also, the schema parameter is
interpreted as a regular expression, so multiple schemas can also be selected
by writing wildcard characters in the pattern. When using wildcards, be careful
to quote the pattern if needed to prevent the shell from expanding the wildcards;
"""
exclude_schema_help_text = """
Do not scan any schemas matching the schema pattern. The pattern is interpreted
according to the same rules as for --include. --exclude can be given more than once to exclude
 schemas matching any of several patterns.

When both --include and ---exclude are given, the behavior is to dump just the schemas that
match at least one --include switch but no --exclude switches.
If --exclude appears without --include, then schemas matching --exclude are excluded from what
is otherwise a normal scan.")
"""
table_help_text = """
Scan only tables matching table. Multiple tables can be selected by writing
multiple switches. Also, the table parameter is interpreted as a regular
expression, so multiple tables can also be selected by writing wildcard
characters in the pattern. When using wildcards, be careful to quote the pattern
 if needed to prevent the shell from expanding the wildcards.
"""
exclude_table_help_text = """
Do not scan any tables matching the table pattern. The pattern is interpreted
according to the same rules as for --include. --exclude can be given more than once to
exclude tables matching any of several patterns.

When both switches are given, the behavior is to dump just the tables that
match at least one --include switch but no --exclude switches. If --exclude appears without
--include, then tables matching --exclude are excluded from what is otherwise a normal scan.
"""


def str_output(op, output_format: OutputFormat):
    if output_format == OutputFormat.tabular:
        return tabulate(
            tabular_data=op,
            headers=("schema", "table", "column", "PII Type", "Scanner"),
        )
    else:
        return json.dumps(op, sort_keys=True, indent=2, cls=PiiTypeEncoder)


@app.command()
def sqlite(
    name: str = typer.Option(..., help="A memorable name for the database"),
    path: Path = typer.Option(..., help="File path to SQLite database"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_sqlite(
            catalog_params=app_state["catalog_connection"],
            name=name,
            path=path,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)


@app.command()
def postgresql(
    name: str = typer.Option(..., help="A memorable name for the database"),
    username: str = typer.Option(..., help="Username or role to connect database"),
    password: str = typer.Option(..., help="Password of username or role"),
    database: str = typer.Option(..., help="Database name"),
    uri: str = typer.Option(..., help="Hostname or URI of the database"),
    port: Optional[int] = typer.Option(None, help="Port number of the database"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_postgresql(
            catalog_params=app_state["catalog_connection"],
            name=name,
            username=username,
            password=password,
            database=database,
            uri=uri,
            port=port,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)


@app.command()
def mysql(
    name: str = typer.Option(..., help="A memorable name for the database"),
    username: str = typer.Option(..., help="Username or role to connect database"),
    password: str = typer.Option(..., help="Password of username or role"),
    database: str = typer.Option(..., help="Database name"),
    uri: str = typer.Option(..., help="Hostname or URI of the database"),
    port: Optional[int] = typer.Option(None, help="Port number of the database"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_mysql(
            catalog_params=app_state["catalog_connection"],
            name=name,
            username=username,
            password=password,
            database=database,
            uri=uri,
            port=port,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)


@app.command()
def redshift(
    name: str = typer.Option(..., help="A memorable name for the database"),
    username: str = typer.Option(..., help="Username or role to connect database"),
    password: str = typer.Option(..., help="Password of username or role"),
    database: str = typer.Option(..., help="Database name"),
    uri: str = typer.Option(..., help="Hostname or URI of the database"),
    port: Optional[int] = typer.Option(None, help="Port number of the database"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_redshift(
            catalog_params=app_state["catalog_connection"],
            name=name,
            username=username,
            password=password,
            database=database,
            uri=uri,
            port=port,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)


@app.command()
def snowflake(
    name: str = typer.Option(..., help="A memorable name for the database"),
    username: str = typer.Option(..., help="Username or role to connect database"),
    password: str = typer.Option(..., help="Password of username or role"),
    database: str = typer.Option(..., help="Database name"),
    account: str = typer.Option(..., help="Snowflake Account Name"),
    warehouse: str = typer.Option(..., help="Snowflake Warehouse Name"),
    role: str = typer.Option(..., help="Snowflake Role Name"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_snowflake(
            catalog_params=app_state["catalog_connection"],
            name=name,
            username=username,
            password=password,
            database=database,
            account=account,
            warehouse=warehouse,
            role=role,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)


@app.command()
def athena(
    name: str = typer.Option(..., help="A memorable name for the database"),
    aws_access_key_id: str = typer.Option(..., help="AWS Access Key"),
    aws_secret_access_key: str = typer.Option(..., help="AWS Secret Key"),
    region_name: str = typer.Option(..., help="AWS Region Name"),
    s3_staging_dir: str = typer.Option(..., help="S3 Staging Dir"),
    scan_type: ScanTypeEnum = typer.Option(
        ScanTypeEnum.shallow,
        help="Choose deep(scan data) or shallow(scan column names only)",
    ),
    incremental: bool = typer.Option(
        True, help="Scan columns updated or created since last run",
    ),
    list_all: bool = typer.Option(
        False,
        help="List all columns. By default only columns with PII information is listed",
    ),
    include_schema: Optional[List[str]] = typer.Option(None, help=schema_help_text),
    exclude_schema: Optional[List[str]] = typer.Option(
        None, help=exclude_schema_help_text
    ),
    include_table: Optional[List[str]] = typer.Option(None, help=table_help_text),
    exclude_table: Optional[List[str]] = typer.Option(
        None, help=exclude_table_help_text
    ),
):
    try:
        op = scan_athena(
            catalog_params=app_state["catalog_connection"],
            name=name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=region_name,
            s3_staging_dir=s3_staging_dir,
            scan_type=scan_type,
            incremental=incremental,
            output_format=app_state["output_format"],
            list_all=list_all,
            include_schema_regex=include_schema,
            exclude_schema_regex=exclude_schema,
            include_table_regex=include_table,
            exclude_table_regex=exclude_table,
        )

        typer.echo(message=str_output(op, app_state["output_format"]))
    except NoMatchesError:
        typer.echo(message=NoMatchesError.message)
        typer.Exit(1)



import datetime
import logging
import re
from typing import Generator, List, Optional, Tuple, Type

from dbcat.catalog import Catalog, CatColumn, CatSchema, CatSource, CatTable
from dbcat.generators import table_generator
from sqlalchemy import create_engine, exc

from piicatcher.dbinfo import DbInfo, get_dbinfo

LOGGER = logging.getLogger(__name__)

SMALL_TABLE_MAX = 100


class NoMatchesError(Exception):
    """Raise Exception if schema/table/column generators do not find any matches"""

    message = "No columns were scanned. Ensure include/exclude patterns are correct OR no new columns have been added"


def column_generator(
    catalog: Catalog,
    source: CatSource,
    last_run: Optional[datetime.datetime] = None,
    include_schema_regex_str: List[str] = None,
    exclude_schema_regex_str: List[str] = None,
    include_table_regex_str: List[str] = None,
    exclude_table_regex_str: List[str] = None,
) -> Generator[Tuple[CatSchema, CatTable, CatColumn], None, None]:

    try:
        for schema, table in table_generator(
            catalog=catalog,
            source=source,
            include_schema_regex_str=include_schema_regex_str,
            exclude_schema_regex_str=exclude_schema_regex_str,
            include_table_regex_str=include_table_regex_str,
            exclude_table_regex_str=exclude_table_regex_str,
        ):

            for column in catalog.get_columns_for_table(
                table=table, newer_than=last_run
            ):
                LOGGER.debug(f"Scanning {schema.name}.{table.name}.{column.name}")
                yield schema, table, column
    except StopIteration:
        raise NoMatchesError


def _get_table_count(
    schema: CatSchema, table: CatTable, dbinfo: Type[DbInfo], connection
) -> int:
    count = dbinfo.get_count_query(schema.name, table.name)
    logging.debug("Count Query: %s" % count)

    result = connection.execute(count)
    row = result.fetchone()

    return int(row[0])


def _get_query(
    schema: CatSchema,
    table: CatTable,
    column_list: List[CatColumn],
    dbinfo: Type[DbInfo],
    connection,
    sample_boundary: int = SMALL_TABLE_MAX,
) -> str:
    count = _get_table_count(schema, table, dbinfo, connection)
    LOGGER.debug("No. of rows in {}.{} is {}".format(schema.name, table.name, count))
    column_name_list: List[str] = [col.name for col in column_list]
    query = dbinfo.get_select_query(schema.name, table.name, column_name_list)

    if count > sample_boundary:
        try:
            query = dbinfo.get_sample_query(schema.name, table.name, column_name_list)
            LOGGER.debug("Choosing a SAMPLE query as table size is big")
        except NotImplementedError:
            LOGGER.warning(
                "Sample Row is not implemented for %s" % dbinfo.__class__.__name__
            )
    return query


def _row_generator(
    source: CatSource, schema: CatSchema, table: CatTable, column_list: List[CatColumn]
):
    engine = create_engine(source.conn_string)
    with engine.connect() as conn:
        dbinfo = get_dbinfo(source.source_type)
        query = _get_query(
            schema=schema,
            table=table,
            column_list=column_list,
            dbinfo=dbinfo,
            connection=conn,
        )
        LOGGER.debug(query)
        result = conn.execute(query)
        row = result.fetchone()
        while row is not None:
            yield row
            row = result.fetchone()


def _filter_text_columns(column_list: List[CatColumn]) -> List[CatColumn]:
    data_type_regex = [
        re.compile(exp, re.IGNORECASE) for exp in [".*char.*", ".*text.*"]
    ]

    matched_set = set()
    for regex in data_type_regex:
        matched_set |= set(
            list(filter(lambda m: regex.search(m.data_type) is not None, column_list,))
        )

    LOGGER.debug(f"{len(matched_set)} text columns found")
    return list(matched_set)


def data_generator(
    catalog: Catalog,
    source: CatSource,
    last_run: Optional[datetime.datetime] = None,
    include_schema_regex_str: List[str] = None,
    exclude_schema_regex_str: List[str] = None,
    include_table_regex_str: List[str] = None,
    exclude_table_regex_str: List[str] = None,
) -> Generator[Tuple[CatSchema, CatTable, CatColumn, str], None, None]:

    for schema, table in table_generator(
        catalog=catalog,
        source=source,
        include_schema_regex_str=include_schema_regex_str,
        exclude_schema_regex_str=exclude_schema_regex_str,
        include_table_regex_str=include_table_regex_str,
        exclude_table_regex_str=exclude_table_regex_str,
    ):

        try:
            columns = _filter_text_columns(
                catalog.get_columns_for_table(table=table, newer_than=last_run)
            )
            if len(columns) > 0:
                for row in _row_generator(
                    column_list=columns, schema=schema, table=table, source=source
                ):
                    for col, val in zip(columns, row):
                        yield schema, table, col, val
        except StopIteration:
            raise NoMatchesError
        except exc.SQLAlchemyError as e:
            LOGGER.warning(
                f"Exception when getting data for {schema.name}.{table.name}. Code: {e.code}"
            )



from abc import ABC, abstractmethod
from typing import List, Type


class DbInfo(ABC):
    _query_template = "select {column_list} from {schema_name}.{table_name}"
    _count_query = "select count(*) from {schema_name}.{table_name}"
    _column_escape = '"'

    @classmethod
    def get_count_query(cls, schema_name: str, table_name: str) -> str:
        return cls._count_query.format(schema_name=schema_name, table_name=table_name)

    @classmethod
    def get_select_query(
        cls, schema_name: str, table_name: str, column_list: List[str]
    ) -> str:
        return cls._query_template.format(
            column_list="{col_list}".format(
                col_list=",".join(
                    "{escape}{name}{escape}".format(name=col, escape=cls._column_escape)
                    for col in column_list
                ),
            ),
            schema_name=schema_name,
            table_name=table_name,
        )

    @classmethod
    @abstractmethod
    def get_sample_query(cls, schema_name, table_name, column_list) -> str:
        pass


class Sqlite(DbInfo):
    _query_template = "select {column_list} from {table_name}"
    _count_query = "select count(*) from {table_name}"

    @classmethod
    def get_select_query(
        cls, schema_name: str, table_name: str, column_list: List[str]
    ) -> str:
        return cls._query_template.format(
            column_list='"{0}"'.format('","'.join(col for col in column_list)),
            table_name=table_name,
        )

    @classmethod
    def get_sample_query(cls, schema_name, table_name, column_list) -> str:
        raise NotImplementedError


class MySQL(DbInfo):
    _sample_query_template = (
        "select {column_list} from {schema_name}.{table_name} limit 10"
    )
    _column_escape = "`"

    @classmethod
    def get_sample_query(cls, schema_name, table_name, column_list) -> str:
        return cls._sample_query_template.format(
            column_list="`{0}`".format("`,`".join(col for col in column_list)),
            schema_name=schema_name,
            table_name=table_name,
        )


class Postgres(DbInfo):
    _sample_query_template = "select {column_list} from {schema_name}.{table_name} TABLESAMPLE BERNOULLI (10)"

    @classmethod
    def get_sample_query(
        cls, schema_name: str, table_name: str, column_list: List[str]
    ) -> str:
        return cls._sample_query_template.format(
            column_list='"{0}"'.format('","'.join(col for col in column_list)),
            schema_name=schema_name,
            table_name=table_name,
        )


class Redshift(Postgres):
    _sample_query_template = "SELECT {column_list} FROM {schema_name}.{table_name} ORDER BY random() LIMIT 10"


class Snowflake(DbInfo):
    _sample_query_template = "select {column_list} from {schema_name}.{table_name} TABLESAMPLE BERNOULLI (10 ROWS)"

    @classmethod
    def get_sample_query(
        cls, schema_name: str, table_name: str, column_list: List[str]
    ) -> str:
        return cls._sample_query_template.format(
            column_list=",".join(column_list),
            schema_name=schema_name,
            table_name=table_name,
        )


def get_dbinfo(source_type: str) -> Type[DbInfo]:
    if source_type == "sqlite":
        return Sqlite
    elif source_type == "mysql":
        return MySQL
    elif source_type == "postgresql":
        return Postgres
    elif source_type == "redshift":
        return Redshift
    elif source_type == "snowflake":
        return Snowflake
    raise AttributeError



import datetime
import logging
from contextlib import closing
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from dbcat.api import init_db, open_catalog
from dbcat.catalog import Catalog, CatSource
from dbcat.catalog.db import DbScanner
from sqlalchemy.orm.exc import NoResultFound

from piicatcher.generators import NoMatchesError, column_generator, data_generator
from piicatcher.output import output_dict, output_tabular
from piicatcher.scanner import deep_scan, shallow_scan

LOGGER = logging.getLogger(__name__)


class ScanTypeEnum(str, Enum):
    shallow = "shallow"
    deep = "deep"


class OutputFormat(str, Enum):
    tabular = "tabular"
    json = "json"


def scan_database(
    catalog: Catalog,
    source: CatSource,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    message = "Source: {source_name}, scan_type: {scan_type}, include_schema: {include_schema}, \
            exclude_schema: {exclude_schema}, include_table: {include_table}, exclude_schema: {exclude_table}".format(
        source_name=source.name,
        scan_type=str(scan_type),
        include_schema=",".join(include_schema_regex)
        if include_schema_regex is not None
        else "None",
        exclude_schema=",".join(exclude_schema_regex)
        if exclude_schema_regex is not None
        else "None",
        include_table=",".join(include_table_regex)
        if include_table_regex is not None
        else "None",
        exclude_table=",".join(exclude_table_regex)
        if exclude_table_regex is not None
        else "None",
    )

    status_message = "Success"
    exit_code = 0

    with catalog.managed_session:
        last_run: Optional[datetime.datetime] = None
        if incremental:
            last_task = catalog.get_latest_task("piicatcher.{}".format(source.name))
            last_run = last_task.updated_at if last_task is not None else None

        try:
            scanner = DbScanner(
                catalog=catalog,
                source=source,
                include_schema_regex_str=include_schema_regex,
                exclude_schema_regex_str=exclude_schema_regex,
                include_table_regex_str=include_table_regex,
                exclude_table_regex_str=exclude_table_regex,
            )
            try:
                scanner.scan()
            except StopIteration:
                raise NoMatchesError

            if scan_type == ScanTypeEnum.shallow:
                shallow_scan(
                    catalog=catalog,
                    generator=column_generator(
                        catalog=catalog,
                        source=source,
                        last_run=last_run,
                        exclude_schema_regex_str=exclude_schema_regex,
                        include_schema_regex_str=include_schema_regex,
                        exclude_table_regex_str=exclude_table_regex,
                        include_table_regex_str=include_table_regex,
                    ),
                )
            else:
                deep_scan(
                    catalog=catalog,
                    generator=data_generator(
                        catalog=catalog,
                        source=source,
                        last_run=last_run,
                        exclude_schema_regex_str=exclude_schema_regex,
                        include_schema_regex_str=include_schema_regex,
                        exclude_table_regex_str=exclude_table_regex,
                        include_table_regex_str=include_table_regex,
                    ),
                )

            if output_format == OutputFormat.tabular:
                return output_tabular(
                    catalog=catalog, source=source, list_all=list_all, last_run=last_run
                )
            else:
                return output_dict(
                    catalog=catalog, source=source, list_all=list_all, last_run=last_run
                )
        except Exception as e:
            status_message = str(e)
            exit_code = 1
            raise e
        finally:
            catalog.add_task(
                "piicatcher.{}".format(source.name),
                exit_code,
                "{}.{}".format(message, status_message),
            )


def scan_sqlite(
    catalog_params: Dict[str, Any],
    name: str,
    path: Path,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
            except NoResultFound:
                source = catalog.add_source(
                    name=path.name, uri=str(path), source_type="sqlite"
                )

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )


def scan_postgresql(
    catalog_params: Dict[str, Any],
    name: str,
    username: str,
    password: str,
    database: str,
    uri: str,
    port: Optional[int] = None,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
            except NoResultFound:
                source = catalog.add_source(
                    name=name,
                    username=username,
                    password=password,
                    database=database,
                    uri=uri,
                    port=port,
                    source_type="postgresql",
                )

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )


def scan_mysql(
    catalog_params: Dict[str, Any],
    name: str,
    username: str,
    password: str,
    database: str,
    uri: str,
    port: Optional[int] = None,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
                LOGGER.debug(f"{name} source found")
            except NoResultFound:
                source = catalog.add_source(
                    name=name,
                    username=username,
                    password=password,
                    database=database,
                    uri=uri,
                    port=port,
                    source_type="mysql",
                )
                LOGGER.debug(f"{name} source added")

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )


def scan_redshift(
    catalog_params: Dict[str, Any],
    name: str,
    username: str,
    password: str,
    database: str,
    uri: str,
    port: Optional[int] = None,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
            except NoResultFound:
                source = catalog.add_source(
                    name=name,
                    username=username,
                    password=password,
                    database=database,
                    uri=uri,
                    port=port,
                    source_type="redshift",
                )

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )


def scan_snowflake(
    catalog_params: Dict[str, Any],
    name: str,
    account: str,
    username: str,
    password: str,
    database: str,
    warehouse: str,
    role: str,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
            except NoResultFound:
                source = catalog.add_source(
                    name=name,
                    username=username,
                    password=password,
                    database=database,
                    account=account,
                    warehouse=warehouse,
                    role=role,
                    source_type="snowflake",
                )

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )


def scan_athena(
    catalog_params: Dict[str, Any],
    name: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    region_name: str,
    s3_staging_dir: str,
    scan_type: ScanTypeEnum = ScanTypeEnum.shallow,
    incremental: bool = True,
    output_format: OutputFormat = OutputFormat.tabular,
    list_all: bool = False,
    include_schema_regex: List[str] = None,
    exclude_schema_regex: List[str] = None,
    include_table_regex: List[str] = None,
    exclude_table_regex: List[str] = None,
) -> Union[List[Any], Dict[Any, Any]]:
    catalog = open_catalog(**catalog_params)

    with closing(catalog) as catalog:
        init_db(catalog)

        with catalog.managed_session:
            try:
                source = catalog.get_source(name)
            except NoResultFound:
                source = catalog.add_source(
                    name=name,
                    aws_access_key_id=aws_access_key_id,
                    aws_secret_access_key=aws_secret_access_key,
                    region_name=region_name,
                    s3_staging_dir=s3_staging_dir,
                    source_type="athena",
                )

            return scan_database(
                catalog=catalog,
                source=source,
                scan_type=scan_type,
                incremental=incremental,
                output_format=output_format,
                list_all=list_all,
                include_schema_regex=include_schema_regex,
                exclude_schema_regex=exclude_schema_regex,
                include_table_regex=include_table_regex,
                exclude_table_regex=exclude_table_regex,
            )



"""Different types of scanners for PII data"""
import logging
import re
from abc import ABC, abstractmethod
from typing import Generator, Tuple

import spacy
from commonregex import CommonRegex
from dbcat.catalog import Catalog
from dbcat.catalog.models import CatColumn, CatSchema, CatTable, PiiTypes

LOGGER = logging.getLogger(__name__)


data_logger = logging.getLogger("piicatcher.data")
data_logger.propagate = False
data_logger.setLevel(logging.INFO)
data_logger.addHandler(logging.NullHandler())

scan_logger = logging.getLogger("piicatcher.scan")
scan_logger.propagate = False
scan_logger.setLevel(logging.INFO)
scan_logger.addHandler(logging.NullHandler())


class Scanner(ABC):
    """Scanner abstract class that defines required methods"""

    @abstractmethod
    def scan(self, text):
        """Scan the text and return an array of PiiTypes that are found"""

    @abstractmethod
    def name(self) -> str:
        """Return name of the scanner"""


class RegexScanner(Scanner):
    """A scanner that uses common regular expressions to find PII"""

    def name(self) -> str:
        return "Regex Scanner on data"

    def scan(self, text):
        """Scan the text and return an array of PiiTypes that are found"""
        regex_result = CommonRegex(text)

        types = []
        if regex_result.phones:  # pylint: disable=no-member
            types.append(PiiTypes.PHONE)
        if regex_result.emails:  # pylint: disable=no-member
            types.append(PiiTypes.EMAIL)
        if regex_result.credit_cards:  # pylint: disable=no-member
            types.append(PiiTypes.CREDIT_CARD)
        if regex_result.street_addresses:  # pylint: disable=no-member
            types.append(PiiTypes.ADDRESS)

        return types


class NERScanner(Scanner):
    """A scanner that uses Spacy NER for entity recognition"""

    def name(self) -> str:
        return "NLP Scanner on data"

    def __init__(self):
        self.nlp = spacy.load("en_core_web_sm")

    def scan(self, text):
        """Scan the text and return an array of PiiTypes that are found"""
        logging.debug("Processing '%s'", text)
        doc = self.nlp(text)
        types = set()
        for ent in doc.ents:
            logging.debug("Found %s", ent.label_)
            if ent.label_ == "PERSON":
                types.add(PiiTypes.PERSON)

            if ent.label_ == "GPE":
                types.add(PiiTypes.LOCATION)

            if ent.label_ == "DATE":
                types.add(PiiTypes.BIRTH_DATE)

        logging.debug("PiiTypes are %s", ",".join(str(x) for x in list(types)))
        return list(types)


class ColumnNameScanner(Scanner):
    def name(self) -> str:
        return "Regular Expression Scanner on column name"

    regex = {
        PiiTypes.PERSON: re.compile(
            "^.*(firstname|fname|lastname|lname|"
            "fullname|maidenname|_name|"
            "nickname|name_suffix|name).*$",
            re.IGNORECASE,
        ),
        PiiTypes.EMAIL: re.compile("^.*(email|e-mail|mail).*$", re.IGNORECASE),
        PiiTypes.BIRTH_DATE: re.compile(
            "^.*(date_of_birth|dateofbirth|dob|"
            "birthday|date_of_death|dateofdeath).*$",
            re.IGNORECASE,
        ),
        PiiTypes.GENDER: re.compile("^.*(gender).*$", re.IGNORECASE),
        PiiTypes.NATIONALITY: re.compile("^.*(nationality).*$", re.IGNORECASE),
        PiiTypes.ADDRESS: re.compile(
            "^.*(address|city|state|county|country|zipcode|postal|zone|borough).*$",
            re.IGNORECASE,
        ),
        PiiTypes.USER_NAME: re.compile("^.*user(id|name|).*$", re.IGNORECASE),
        PiiTypes.PASSWORD: re.compile("^.*pass.*$", re.IGNORECASE),
        PiiTypes.SSN: re.compile("^.*(ssn|social).*$", re.IGNORECASE),
    }

    def scan(self, text):
        types = set()
        for pii_type in self.regex:
            if self.regex[pii_type].match(text) is not None:
                types.add(pii_type)

        return list(types)


def shallow_scan(
    catalog: Catalog,
    generator: Generator[Tuple[CatSchema, CatTable, CatColumn], None, None],
):
    scanner = ColumnNameScanner()

    for schema, table, column in generator:
        types = scanner.scan(column.name)
        if len(types) > 0:
            catalog.set_column_pii_type(
                column=column, pii_type=types.pop(), pii_plugin=scanner.name()
            )


def deep_scan(
    catalog: Catalog,
    generator: Generator[Tuple[CatSchema, CatTable, CatColumn, str], None, None],
):
    scanners = [RegexScanner(), NERScanner()]

    for schema, table, column, val in generator:
        LOGGER.debug("Scanning column name {}, val: {}".format(column.fqdn, val))
        if val is not None:
            pii_types = set()
            for scanner in scanners:
                for pii in scanner.scan(val):
                    pii_types.add(pii)

                if len(pii_types) > 0:
                    top = pii_types.pop()
                    catalog.set_column_pii_type(
                        column=column, pii_type=top, pii_plugin=scanner.name()
                    )
                    LOGGER.debug("{} has {}".format(column.fqdn, top))

                    scan_logger.info(
                        "deep_scan", extra={"column": column.fqdn, "pii_types": top}
                    )
                    data_logger.info(
                        "deep_scan",
                        extra={"column": column.fqdn, "data": val, "pii_types": top},
                    )



import logging
import logging.config
from pathlib import Path
from typing import Optional

import typer
from pythonjsonlogger import jsonlogger

from piicatcher import __version__
from piicatcher.api import OutputFormat
from piicatcher.app_state import app_state
from piicatcher.cli import app as scan_app
from piicatcher.scanner import data_logger, scan_logger

app = typer.Typer()

LOGGER = logging.getLogger(__name__)


class TyperLoggerHandler(logging.Handler):
    def emit(self, record: logging.LogRecord) -> None:
        fg = None
        bg = None
        if record.levelno == logging.DEBUG:
            fg = typer.colors.YELLOW
        elif record.levelno == logging.INFO:
            fg = typer.colors.BRIGHT_BLUE
        elif record.levelno == logging.WARNING:
            fg = typer.colors.BRIGHT_MAGENTA
        elif record.levelno == logging.CRITICAL:
            fg = typer.colors.BRIGHT_RED
        elif record.levelno == logging.ERROR:
            fg = typer.colors.BRIGHT_WHITE
            bg = typer.colors.RED
        typer.secho(self.format(record), bg=bg, fg=fg)


def log_config(log_level: str):
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {
                "format": "{asctime}:{levelname}:{name} {message}",
                "style": "{",
            },
        },
        "handlers": {
            "console": {"class": "logging.StreamHandler", "formatter": "simple",},
            "typer": {
                "class": "piicatcher.command_line.TyperLoggerHandler",
                "formatter": "simple",
            },
        },
        "root": {"handlers": ["typer"], "level": log_level},
        "piicatcher.api": {"handlers": ["typer"], "level": log_level},
    }


def version_callback(value: bool):
    if value:
        print("{}".format(__version__))
        typer.Exit()


# pylint: disable=too-many-arguments
@app.callback()
def cli(
    log_level: str = typer.Option("WARNING", help="Logging Level"),
    log_data: bool = typer.Option(False, help="Choose output format type"),
    log_scan: bool = typer.Option(False, help="Log data that was scanned"),
    output_format: OutputFormat = typer.Option(
        OutputFormat.tabular, case_sensitive=False
    ),
    catalog_path: Path = typer.Option(
        None, help="Path to store catalog state. Use if NOT using a database"
    ),
    catalog_host: str = typer.Option(
        None, help="hostname of Postgres database. Use if catalog is a database."
    ),
    catalog_port: int = typer.Option(
        None, help="port of Postgres database. Use if catalog is a database."
    ),
    catalog_user: str = typer.Option(
        None, help="user of Postgres database. Use if catalog is a database."
    ),
    catalog_password: str = typer.Option(
        None, help="password of Postgres database. Use if catalog is a database."
    ),
    catalog_database: str = typer.Option(
        None, help="database of Postgres database. Use if catalog is a database."
    ),
    version: Optional[bool] = typer.Option(
        None, "--version", callback=version_callback
    ),
):
    logging.config.dictConfig(log_config(log_level=log_level.upper()))

    if log_scan:
        handler = logging.StreamHandler()
        handler.setFormatter(jsonlogger.JsonFormatter())
        handler.setLevel(logging.INFO)
        scan_logger.addHandler(handler)
        LOGGER.debug("SCAN LOG setup")

    if log_data:
        handler = logging.StreamHandler()
        handler.setFormatter(jsonlogger.JsonFormatter())
        handler.setLevel(logging.INFO)
        data_logger.addHandler(handler)
        LOGGER.debug("DATA LOG setup")

    app_dir = typer.get_app_dir("tokern")
    app_dir_path = Path(app_dir)
    app_dir_path.mkdir(parents=True, exist_ok=True)

    app_state["catalog_connection"] = {
        "path": str(catalog_path),
        "user": catalog_user,
        "password": catalog_password,
        "host": catalog_host,
        "port": catalog_port,
        "database": catalog_database,
        "app_dir": app_dir,
    }
    app_state["output_format"] = output_format


app.add_typer(scan_app, name="scan")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['piicatcher']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'commonregex>=1.5,<2.0',
 'dbcat>=0.10.0,<0.11.0',
 'python-json-logger>=2.0.2,<3.0.0',
 'pyyaml',
 'spacy',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.4.0,<0.5.0']

extras_require = \
{'datahub': ['great-expectations>=0.13.42,<0.14.0',
             'acryl-datahub>=0.8.16,<0.9.0']}

entry_points = \
{'console_scripts': ['piicatcher = piicatcher.command_line:app']}

setup_kwargs = {
    'name': 'piicatcher',
    'version': '0.17.0',
    'description': 'Find PII data in databases',
    'long_description': '[![CircleCI](https://circleci.com/gh/tokern/piicatcher.svg?style=svg)](https://circleci.com/gh/tokern/piicatcher)\n[![codecov](https://codecov.io/gh/tokern/piicatcher/branch/master/graph/badge.svg)](https://codecov.io/gh/tokern/piicatcher)\n[![PyPI](https://img.shields.io/pypi/v/piicatcher.svg)](https://pypi.python.org/pypi/piicatcher)\n[![image](https://img.shields.io/pypi/l/piicatcher.svg)](https://pypi.org/project/piicatcher/)\n[![image](https://img.shields.io/pypi/pyversions/piicatcher.svg)](https://pypi.org/project/piicatcher/)\n[![image](https://img.shields.io/docker/v/tokern/piicatcher)](https://hub.docker.com/r/tokern/piicatcher)\n\n# PII Catcher for Databases and Data Warehouses\n\n## Overview\n\nPIICatcher is a data catalog and scanner for PII and PHI information. It finds PII data in your databases and file systems\nand tracks critical data. The data catalog can be used as a foundation to build governance, compliance and security\napplications.\n\nCheck out [AWS Glue & Lake Formation Privilege Analyzer](https://tokern.io/blog/lake-glue-access-analyzer) for an example of how piicatcher is used in production.\n\n## Quick Start\n\nPIICatcher is available as a docker image or command-line application.\n\n### Docker\n\n    docker run tokern/piicatcher:latest scan sqlite --path \'/db/sqlqb\'\n\n    ╭─────────────┬─────────────┬─────────────┬─────────────╮\n    │   schema    │    table    │   column    │   has_pii   │\n    ├─────────────┼─────────────┼─────────────┼─────────────┤\n    │        main │    full_pii │           a │           1 │\n    │        main │    full_pii │           b │           1 │\n    │        main │      no_pii │           a │           0 │\n    │        main │      no_pii │           b │           0 │\n    │        main │ partial_pii │           a │           1 │\n    │        main │ partial_pii │           b │           0 │\n    ╰─────────────┴─────────────┴─────────────┴─────────────╯\n\n### Command-line\nTo install use pip:\n\n    python3 -m venv .env\n    source .env/bin/activate\n    pip install piicatcher\n\n    # Install Spacy English package\n    python -m spacy download en_core_web_sm\n    \n    # run piicatcher on a sqlite db and print report to console\n    piicatcher scan sqlite --path \'/db/sqlqb\'\n    ╭─────────────┬─────────────┬─────────────┬─────────────╮\n    │   schema    │    table    │   column    │   has_pii   │\n    ├─────────────┼─────────────┼─────────────┼─────────────┤\n    │        main │    full_pii │           a │           1 │\n    │        main │    full_pii │           b │           1 │\n    │        main │      no_pii │           a │           0 │\n    │        main │      no_pii │           b │           0 │\n    │        main │ partial_pii │           a │           1 │\n    │        main │ partial_pii │           b │           0 │\n    ╰─────────────┴─────────────┴─────────────┴─────────────╯\n\n\n### API\n    from piicatcher.api import scan_postgresql\n\n    # PIICatcher uses a catalog to store its state. \n    # The easiest option is to use a sqlite memory database.\n    # For production usage check, https://tokern.io/docs/data-catalog\n    catalog_params={\'catalog_path\': \':memory:\'}\n    output = scan_postrgresql(catalog_params=catalog_params, name="pg_db", uri="127.0.0.1", \n                              username="piiuser", password="p11secret", database="piidb", \n                              include_table_regex=["sample"])\n    print(output)\n\n    # Example Output\n    [[\'public\', \'sample\', \'gender\', \'PiiTypes.GENDER\'], \n     [\'public\', \'sample\', \'maiden_name\', \'PiiTypes.PERSON\'], \n     [\'public\', \'sample\', \'lname\', \'PiiTypes.PERSON\'], \n     [\'public\', \'sample\', \'fname\', \'PiiTypes.PERSON\'], \n     [\'public\', \'sample\', \'address\', \'PiiTypes.ADDRESS\'], \n     [\'public\', \'sample\', \'city\', \'PiiTypes.ADDRESS\'], \n     [\'public\', \'sample\', \'state\', \'PiiTypes.ADDRESS\'], \n     [\'public\', \'sample\', \'email\', \'PiiTypes.EMAIL\']]\n\n\n## Supported Databases\n\nPIICatcher supports the following databases:\n1. **Sqlite3** v3.24.0 or greater\n2. **MySQL** 5.6 or greater\n3. **PostgreSQL** 9.4 or greater\n4. **AWS Redshift**\n5. **AWS Athena**\n6. **Snowflake**\n\n## Documentation\n\nFor advanced usage refer documentation [PIICatcher Documentation](https://tokern.io/docs/piicatcher).\n\n## Survey\n\nPlease take this [survey](https://forms.gle/Ns6QSNvfj3Pr2s9s6) if you are a user or considering using PIICatcher. \nThe responses will help to prioritize improvements to the project.\n\n## Contributing\n\nFor Contribution guidelines, [PIICatcher Developer documentation](https://tokern.io/docs/piicatcher/development). \n\n',
    'author': 'Tokern',
    'author_email': 'info@tokern.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://tokern.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
